/*
 * hid.c - UZK (USB Keyboard Bridge) Host Controller
 *   by uratan! 2004.5.28
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

#define INPUT_REPORT_SIZE  2		/* LED status */
#define OUTPUT_REPORT_SIZE  9		/* keycodes */

int hid_read(u_char *buff);
void clear_k_m_memory(void);
void modify_usb_keycodes(int to_add, u_char id);
int hid_write(u_char *buff);

int uhid;

u_char usb_led[INPUT_REPORT_SIZE];

u_char key_data[OUTPUT_REPORT_SIZE];
u_char mouse_data[OUTPUT_REPORT_SIZE];		/* not used */

/* in xps2k.c */
extern int verbose;


/*
 * read input report for LED status
 */
int hid_read(u_char *buff)
{
	int c, i;
	fd_set fds;
	struct timeval timeout;

	FD_ZERO(&fds);
	FD_SET(uhid, &fds);
	timeout.tv_sec = 0;
	timeout.tv_usec = 1*1000;	/* 1msec */

	c = select(uhid+1, &fds, NULL, NULL, &timeout);
	if(c < 0) {
		/*
		 * fail
		 */
		fprintf(stderr, "fail to select()\n");
		return(c);
	}

	if(c == 0) {
		/*
		 * not ready for read
		 */
		return(c);
	}

	/*
	 * read LED status
	 */
	c = read(uhid, buff, INPUT_REPORT_SIZE);

	if(c < 0) {
		fprintf(stderr, "can't read() from HID device\n");
		return(c);
	}

	if(verbose) {
		printf("input report:");
		for(i=0; i<INPUT_REPORT_SIZE; i++)
			printf(" %02x", buff[i]);
		printf(" has come\n");
	}

	return(c);
}

/*
 * initial key and mouse report data
 */
void clear_k_m_memory(void)
{
	int i;

	/*
	 * fill all by 0
	 */
	for(i=0; i<OUTPUT_REPORT_SIZE; i++) {
		key_data[i] = 0x00;
		mouse_data[i] = 0x00;
	}

	/*
	 * set header
	 */
	key_data[0] = 'K';
	mouse_data[0] = 'M';

	return;
}

/*
 * add/remove a keycode to/from key_data[]
 */
void modify_usb_keycodes(int to_add, u_char id)
{
	int i;

	if(to_add) {
		for(i=3; i<OUTPUT_REPORT_SIZE; i++) {
			if(key_data[i] == 0x00) {
				key_data[i] = id;
				break;		/* add only 1 */
			}
		}
	} else {
		for(i=3; i<OUTPUT_REPORT_SIZE; i++) {
			if(key_data[i] == id)
				key_data[i] = 0x00;	/* remove all */
		}
	}

	return;
}

/*
 * write output report for Keycodes emulation
 */
int hid_write(u_char *buff)
{
	int i, c;

	if(verbose) {
		printf("output report:");
		for(i=0; i<OUTPUT_REPORT_SIZE; i++)
			printf(" %02x", buff[i]);
		printf(" ... ");
	}

	c = write(uhid, buff, OUTPUT_REPORT_SIZE);
	if(c < 0) {
		fprintf(stderr, "can't write() for HID device\n");
		return(c);
	}

	if(verbose)
		printf("succeed\n");

	return(c);
}
