/*
 * keyconv.c - convert keysyms for ps2 scancode
 *   by uratan! 2003.12.1
 */
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>


/*
 * my local keysyms for cur & paste (char size)
 */
#define xk_shift    0x00e		/* local SHIFT */
#define xk_return   0x00a		/* local return */

/*
 * structure for conversion table
 */
typedef struct {
	int    sc_mode;		/* scancode modifier */
	int    sc_code;		/* scancode itself  */
#ifdef HID_UZK
	int    id;			/* usage ID (key code on USB keyboard */
#endif /* HID_UZK */
	KeySym ks1;			/* KeySym for normal */
	KeySym ks2;			/* KeySym for SHIFT */
} T_SCANCODE_FROM_KEYSYM;


/*
 * USB modifier keys
 */
#define U_L_C     (1<< 8)	/* Left Control */
#define U_L_S     (1<< 9)	/* Left Shift */
#define U_L_A     (1<<10)	/* Left Alt */
#define U_L_G     (1<<11)	/* Left Gui */
#define U_R_C     (1<<12)	/* Right Control */
#define U_R_S     (1<<13)	/* Right Shift */
#define U_R_A     (1<<14)	/* Right Alt */
#define U_R_G     (1<<15)	/* Right Gui */

/*
 * scancode modifiers
 */
#define NORM  1		/* code / 0xf0+code */
#define S_E0  2		/* 0xe0+code / 0xe0+0xf0+code */
#define PRNT  3		/* special */
#define PAUS  4		/* special */


/*
 * prototypes
 */
int convert_and_send_C_A(int pressed);
int convert_and_send(KeySym keysym, int pressed);
char *make_key_strokes(char *m);
const T_SCANCODE_FROM_KEYSYM scancode_from_single_keysym[];
const T_SCANCODE_FROM_KEYSYM scancode_from_double_keysym101[];


/*
 * externs
 */
extern int verbose;							/* in xps2k.c */

#ifdef HID_UZK
/* in hid.c */
extern void modify_usb_keycodes(int to_add, u_char id);
extern int hid_write(u_char *buff);
extern u_char key_data[];
#else /* HID_UZK */
extern int try_to_send_a_byte(int byte);	/* in ps2.c */
#endif /* HID_UZK */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * convert CTRL and ALT to ps2 scancode, then send it
 */
int convert_and_send_C_A(int pressed)
{
	int c;

	c = convert_and_send(XK_Control_R, pressed);
	if(c < 0)
		return(c);
	c = convert_and_send(XK_Alt_R, pressed);
	if(c < 0)
		return(c);

	return(0);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * convert keycode to ps2 scancode, then send it
 */
int convert_and_send(KeySym keysym, int pressed)
{
	const T_SCANCODE_FROM_KEYSYM *table;
	int i, len, c;
#ifndef HID_UZK
	int scancode[10];
#endif /* !HID_UZK */

	/*
	 * check the parameter
	 */
	if(keysym == NoSymbol) {
		if(verbose)
			printf(" no symbol assinged\n");
		return(0);		/* need nothing for success */
	}

	/*
	 * override these for cut & paste
	 */
	if(keysym == xk_shift) {
		/*
		 * SHIFT
		 */
		keysym = XK_Shift_R;
	} else if(keysym == xk_return) {
		/*
		 * only replace keysym
		 */
		keysym = XK_Return;
	}

	/*
	 * select a table
	 */
	if(keysym >= 0x100)
		table = scancode_from_single_keysym;
	else
		table = scancode_from_double_keysym101;

	/*
	 * search specified KeySym in the table
	 */
	while(1) {
		if(table->ks1 == -1) {
			if(verbose)
				printf(" not found in the table\n");
			return(0);		/* reach to end of the table */
		}

		if((table->ks1 == keysym) || (table->ks2 == keysym))
			break;

		table++;
	}

	/*
	 * generate scancode
	 */
#ifdef HID_UZK
	/*
	 * UZK on USB
	 */
	if(table->id >= 0x100) {
		if(pressed)
			key_data[1] |= table->id >> 8;
		else
			key_data[1] &= ~table->id >> 8;
	} else {
		modify_usb_keycodes(pressed, table->id);
	}

	c = hid_write(key_data);
	if(c < 0)
		return(c);
#else /* HID_UZK */
	/*
	 * EZK on tty
	 */
	switch(table->sc_mode) {
	case NORM:
			if(pressed) {
				len = 1;
				scancode[0] = table->sc_code;
			} else /* released */ {
				len = 2;
				scancode[0] = 0xf0;
				scancode[1] = table->sc_code;
			}
			break;

	case S_E0:
			if(pressed) {
				len = 2;
				scancode[0] = 0xe0;
				scancode[1] = table->sc_code;
			} else /* released */ {
				len = 3;
				scancode[0] = 0xe0;
				scancode[1] = 0xf0;
				scancode[2] = table->sc_code;
			}
			break;

	case PRNT:
			if(pressed) {
				len = 4;
				scancode[0] = 0xe0;
				scancode[1] = 0x12;
				scancode[2] = 0xe0;
				scancode[3] = 0x7c;
			} else /* released */ {
				len = 6;
				scancode[0] = 0xe0;
				scancode[1] = 0xf0;
				scancode[2] = 0x7c;
				scancode[3] = 0xe0;
				scancode[4] = 0xf0;
				scancode[5] = 0x12;
			}
			break;

	case PAUS:
			if(pressed) {
				len = 8;
				scancode[0] = 0xe1;
				scancode[1] = 0x14;
				scancode[2] = 0x77;
				scancode[3] = 0xe1;
				scancode[4] = 0xf0;
				scancode[5] = 0x14;
				scancode[6] = 0xf0;
				scancode[7] = 0x77;
			} else /* released */ {
				len = 0;
				/* nothing to send */
			}
			break;
	}

	/*
	 * display the scancode if verbose mode
	 */
	if(verbose) {
		printf(" --> scancode");
		if(len!=0) {
			for(i=0; i<len; i++)
				printf(" %02x", scancode[i]);
		} else {
			printf(" (none)");
		}
		printf("\n");
		fflush(stdout);
	}

	/*
	 * send the scancode
	 */
	for(i=0; i<len; i++) {
		c = try_to_send_a_byte(scancode[i]);
		if(c < 0)
			return(c);
	}
#endif /* HID_UZK */

	return(0);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * convert normal string for key stroke
 *
 *   "http://w...
 *
 *     ==> 'h', press, 'h', release
 *         't', press, 't', release
 *         't', press, 't', release
 *         'p', press, 'p', release
 *         [S], press,
 *         ':', press, ':', release
 *         [S], release,
 *         '/', press '/', release
 *         '/', press '/', release
 *         'w', press 'w', release
 *         ...
 */
char *make_key_strokes(char *s)
{
	int len, i, c;
	char *buff, *t;
	int shift_is_on, need_shift;
	const T_SCANCODE_FROM_KEYSYM *table;

	if(s == NULL)
		return(NULL);
		
	len = strlen(s);
	if(len == 0)
		return(NULL);

	buff = (char *)malloc(2 + len * 6 + 4);
	if(buff == NULL)
		return(NULL);

	t = buff;
	shift_is_on = 0;
	*t++ = xk_shift;
	*t++ = 0;		/* released */

	for(i=0; i<len; i++) {
		c = s[i];

		if(c == '\n') {
			*t++ = xk_return;
			*t++ = 1;		/* pressed */
			*t++ = xk_return;
			*t++ = 0;		/* released */
			continue;
		}

		if((c < ' ') || ('~' < c))
			continue;

		table = scancode_from_double_keysym101;
		while(1) {
			if(table->ks1 == -1)
				goto next_char;		/* reach to end of the table, ignore */

			if(table->ks1 == c) {
				need_shift = 0;
				break;
			} else if(table->ks2 == c) {
				need_shift = 1;
				break;
			}

			table++;
		}

		if((shift_is_on==1) && (need_shift==0)) {
			*t++ = xk_shift;
			*t++ = 0;		/* released */
			shift_is_on = 0;
		} else if((shift_is_on==0) && (need_shift==1)) {
			*t++ = xk_shift;
			*t++ = 1;		/* pressed */
			shift_is_on = 1;
		}
		*t++ = c;	/* add the char */
		*t++ = 1;	/* pressed */
		*t++ = c;	/* add the char */
		*t++ = 0;	/* released */

next_char:
	}

	*t++ = xk_shift;	/* add shift */
	*t++ = 0;			/* released */

	*t++ = '\0';		/* set end of strokes */
	*t++ = '\0';		/* set end of strokes */

	return(buff);
}

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * keysym to scancode table when keysym >= 0x100
 *   these may be common between 101/106 keyboards
 */
const T_SCANCODE_FROM_KEYSYM scancode_from_single_keysym[] = {
/*(ESC)*/	{ NORM, 0x76,	0x29,	XK_Escape,		NoSymbol		},
/*(TAB)*/	{ NORM, 0x0d,	0x2b,	XK_Tab,			NoSymbol		},
/*(BS)*/	{ NORM, 0x66,	0x2a,	XK_BackSpace,	NoSymbol		},
/*(CTLL)*/	{ NORM, 0x14,	U_L_C,	XK_Control_L,	NoSymbol		},
/*(CTLR)*/	{ S_E0, 0x14,	U_R_C,	XK_Control_R,	NoSymbol		},
/*(RET)*/	{ NORM, 0x5a,	0x28,	XK_Return,		NoSymbol		},
/*(SHTL)*/	{ NORM, 0x12,	U_L_S,	XK_Shift_L,		NoSymbol		},
/*(SHTR)*/	{ NORM, 0x59,	U_R_S,	XK_Shift_R,		NoSymbol		},
/*(ALTL)*/	{ NORM, 0x11,	U_L_A,	XK_Alt_L,		NoSymbol		},
/*(ALTR)*/	{ S_E0, 0x11,	U_R_A,	XK_Alt_R,		NoSymbol		},
/*(METAL)*/	{ S_E0, 0x1f,	U_L_G,	XK_Meta_L,		NoSymbol		},
/*(METAR)*/	{ S_E0, 0x27,	U_R_G,	XK_Meta_R,		NoSymbol		},
/*(UP)*/	{ S_E0, 0x75,	0x52,	XK_Up,			NoSymbol		},
/*(<--)*/	{ S_E0, 0x6b,	0x50,	XK_Left,		NoSymbol		},
/*(DOWN)*/	{ S_E0, 0x72,	0x51,	XK_Down,		NoSymbol		},
/*(-->)*/	{ S_E0, 0x74,	0x4f,	XK_Right,		NoSymbol		},
/*(F1)*/	{ NORM, 0x05,	0x3a,	XK_F1,			NoSymbol		},
/*(F2)*/	{ NORM, 0x06,	0x3b,	XK_F2,			NoSymbol		},
/*(F3)*/	{ NORM, 0x04,	0x3c,	XK_F3,			NoSymbol		},
/*(F4)*/	{ NORM, 0x0c,	0x3d,	XK_F4,			NoSymbol		},
/*(F5)*/	{ NORM, 0x03,	0x3e,	XK_F5,			NoSymbol		},
/*(F6)*/	{ NORM, 0x0b,	0x3f,	XK_F6,			NoSymbol		},
/*(F7)*/	{ NORM, 0x83,	0x40,	XK_F7,			NoSymbol		},
/*(F8)*/	{ NORM, 0x0a,	0x41,	XK_F8,			NoSymbol		},
/*(F9)*/	{ NORM, 0x01,	0x42,	XK_F9,			NoSymbol		},
/*(F10)*/	{ NORM, 0x09,	0x43,	XK_F10,			NoSymbol		},
/*(F11)*/	{ NORM, 0x78,	0x44,	XK_F11,			NoSymbol		},
/*(F12)*/	{ NORM, 0x07,	0x45,	XK_F12,			NoSymbol		},
/*(Ins)*/	{ S_E0, 0x70,	0x49,	XK_Insert,		NoSymbol		},
/*(Del)*/	{ S_E0, 0x71,	0x4c,	XK_Delete,		NoSymbol		},
/*(PSc)*/	{ PRNT, 0x00,	0x46,	XK_Print,		NoSymbol		},
/*(SRq)*/	{ NORM, 0x84,	0x46,	XK_Sys_Req,		NoSymbol		},/*sameasPrt*/
/*(ScrL)*/	{ NORM, 0x7e,	0x47,	XK_Scroll_Lock,	NoSymbol		},
/*(Pus)*/	{ PAUS, 0x00,	0x48,	XK_Pause,		NoSymbol		},
/*(PgUp)*/	{ S_E0, 0x7d,	0x4b,	XK_Prior,		NoSymbol		},
/*(PgDn)*/	{ S_E0, 0x7a,	0x4e,	XK_Next,		NoSymbol		},
/*(Home)*/	{ S_E0, 0x6c,	0x4a,	XK_Home,		NoSymbol		},
/*(End)*/	{ S_E0, 0x69,	0x4d,	XK_End,			NoSymbol		},
			{ -1,   -1,		-1,		-1,				-1				},
};
#if 0
	/*
	 * scancode for other keys as sample
	 */
	/*(LOCK)*/	{ NORM, 0x58,	0x39,	XK_Caps_Lock?,	NoSymbol		},
	/*(NumL)*/	{ NORM, 0x77,	0x53,	XK_Num_Lock,	NoSymbol		},
	/*(KP_0)*/	{ NORM, 0x70,	0x62,	XK_KP_0,		NoSymbol		},
	/*(KP_1)*/	{ NORM, 0x69,	0x59,	XK_KP_1,		NoSymbol		},
	/*(KP_2)*/	{ NORM, 0x72,	0x5a,	XK_KP_2,		NoSymbol		},
	/*(KP_3)*/	{ NORM, 0x7a,	0x5b,	XK_KP_3,		NoSymbol		},
	/*(KP_4)*/	{ NORM, 0x6b,	0x5c,	XK_KP_4,		NoSymbol		},
	/*(KP_5)*/	{ NORM, 0x73,	0x5d,	XK_KP_5,		NoSymbol		},
	/*(KP_6)*/	{ NORM, 0x74,	0x5e,	XK_KP_6,		NoSymbol		},
	/*(KP_7)*/	{ NORM, 0x6c,	0x5f,	XK_KP_7,		NoSymbol		},
	/*(KP_8)*/	{ NORM, 0x75,	0x60,	XK_KP_8,		NoSymbol		},
	/*(KP_9)*/	{ NORM, 0x7d,	0x61,	XK_KP_9,		NoSymbol		},
	/*(KP_.)*/	{ NORM, 0x71,	0x63,	XK_KP_Decimal,	NoSymbol		},
	/*(KP_-)*/	{ NORM, 0x7b,	0x56,	XK_KP_Subtract,	NoSymbol		},
	/*(KP_+)*/	{ NORM, 0x79,	0x57,	XK_KP_Add,		NoSymbol		},
	/*(KP_*)*/	{ NORM, 0x7c,	0x55,	XK_KP_Multiply,	NoSymbol		},
	/*(KP_/)*/	{ S_E0, 0x4a,	0x54,	XK_KP_Divide,	NoSymbol		},
	/*(KP_E)*/	{ S_E0, 0x5a,	0x58,	XK_KP_Enter,	NoSymbol		},
	/*(WinL)*/	{ S_E0, 0x1f,	0x00,	XK_Windows_L?,	NoSymbol		},
	/*(WinR)*/	{ S_E0, 0x27,	0x00,	XK_Windows_R?,	NoSymbol		},
	/*(App)*/	{ S_E0, 0x2f,	0x00,	XK_Win_app?,	NoSymbol		},
	/*(MUHEN)*/	{ NORM, 0x67,	0x00,	XK_Muhenkan?,	NoSymbol		},
	/*(MAE)*/	{ NORM, 0x64,	0x00,	XK_Mae_Koho?,	NoSymbol		},
	/*(KANA)*/	{ NORM, 0x13,	0x00,	XK_Katakana?,	NoSymbol		},
#endif /* 0 */

/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */

/*
 * keysym to scancode table when keysym < 0x100, for 101 keyboard
 */
const T_SCANCODE_FROM_KEYSYM scancode_from_double_keysym101[] = {
/*( / )*/	{ NORM, 0x29,	0x2c,	XK_space,		XK_space		},
/*(1/!)*/	{ NORM, 0x16,	0x1e,	XK_1,			XK_exclam		},
 /*(2/@)*/	{ NORM, 0x1e,	0x1f,	XK_2,			XK_at			},	/*1*/
/*(3/#)*/	{ NORM, 0x26,	0x20,	XK_3,			XK_numbersign	},
/*(4/$)*/	{ NORM, 0x25,	0x21,	XK_4,			XK_dollar		},
/*(5/%)*/	{ NORM, 0x2e,	0x22,	XK_5,			XK_percent		},
 /*(6/^)*/	{ NORM, 0x36,	0x23,	XK_6,			XK_asciicircum	},	/*1*/
 /*(7/&)*/	{ NORM, 0x3d,	0x24,	XK_7,			XK_ampersand	},	/*1*/
 /*(8/*)*/	{ NORM, 0x3e,	0x25,	XK_8,			XK_asterisk		},	/*1*/
 /*(9/()*/	{ NORM, 0x46,	0x26,	XK_9,			XK_parenleft	},	/*1*/
 /*(0/))*/	{ NORM, 0x45,	0x27,	XK_0,			XK_parenright	},	/*1*/
 /*(-/_)*/	{ NORM, 0x4e,	0x2d,	XK_minus,		XK_underscore	},	/*1*/
 /*(=/+)*/	{ NORM, 0x55,	0x2e,	XK_equal,		XK_plus			},	/*1*/
 /*(\/|)*/	{ NORM, 0x5d,	0x31,	XK_backslash,	XK_bar			},	/*1*/
 /*(`/~)*/	{ NORM, 0x0e,	0x35,	XK_grave,		XK_asciitilde	},	/*1*/
/*(q/Q)*/	{ NORM, 0x15,	0x14,	XK_q,			XK_Q			},
/*(w/W)*/	{ NORM, 0x1d,	0x1a,	XK_w,			XK_W			},
/*(e/E)*/	{ NORM, 0x24,	0x08,	XK_e,			XK_E			},
/*(r/R)*/	{ NORM, 0x2d,	0x15,	XK_r,			XK_R			},
/*(t/T)*/	{ NORM, 0x2c,	0x17,	XK_t,			XK_T			},
/*(y/Y)*/	{ NORM, 0x35,	0x1c,	XK_y,			XK_Y			},
/*(u/U)*/	{ NORM, 0x3c,	0x18,	XK_u,			XK_U			},
/*(i/I)*/	{ NORM, 0x43,	0x0c,	XK_i,			XK_I			},
/*(o/O)*/	{ NORM, 0x44,	0x12,	XK_o,			XK_O			},
/*(p/P)*/	{ NORM, 0x4d,	0x13,	XK_p,			XK_P			},
 /*([/{)*/	{ NORM, 0x54,	0x2f,	XK_bracketleft,	XK_braceleft	},	/*1*/
 /*(]/})*/	{ NORM, 0x5b,	0x30,	XK_bracketright,XK_braceright	},	/*1*/
/*(a/A)*/	{ NORM, 0x1c,	0x04,	XK_a,			XK_A			},
/*(s/S)*/	{ NORM, 0x1b,	0x16,	XK_s,			XK_S			},
/*(d/D)*/	{ NORM, 0x23,	0x07,	XK_d,			XK_D			},
/*(f/F)*/	{ NORM, 0x2b,	0x09,	XK_f,			XK_F			},
/*(g/G)*/	{ NORM, 0x34,	0x0a,	XK_g,			XK_G			},
/*(h/H)*/	{ NORM, 0x33,	0x0b,	XK_h,			XK_H			},
/*(j/J)*/	{ NORM, 0x3b,	0x0d,	XK_j,			XK_J			},
/*(k/K)*/	{ NORM, 0x42,	0x0e,	XK_k,			XK_K			},
/*(l/L)*/	{ NORM, 0x4b,	0x0f,	XK_l,			XK_L			},
 /*(;/:)*/	{ NORM, 0x4c,	0x33,	XK_semicolon,	XK_colon		},	/*1*/
 /*('/")*/	{ NORM, 0x52,	0x34,	XK_apostrophe,	XK_quotedbl		},	/*1*/
/*(z/Z)*/	{ NORM, 0x1a,	0x1d,	XK_z,			XK_Z			},
/*(x/X)*/	{ NORM, 0x22,	0x1b,	XK_x,			XK_X			},
/*(c/C)*/	{ NORM, 0x21,	0x06,	XK_c,			XK_C			},
/*(v/V)*/	{ NORM, 0x2a,	0x19,	XK_v,			XK_V			},
/*(b/B)*/	{ NORM, 0x32,	0x05,	XK_b,			XK_B			},
/*(n/N)*/	{ NORM, 0x31,	0x11,	XK_n,			XK_N			},
/*(m/M)*/	{ NORM, 0x3a,	0x10,	XK_m,			XK_M			},
/*(,/<)*/	{ NORM, 0x41,	0x36,	XK_comma,		XK_less			},
/*(./>)*/	{ NORM, 0x49,	0x37,	XK_period,		XK_greater		},
/*(//?)*/	{ NORM, 0x4a,	0x38,	XK_slash,		XK_question		},
			{ -1,   -1,		-1,		-1,				-1				},
};

#if 0
/*
 * keysym to scancode table when keysym < 0x100, for 106 keyboard
 */
const T_SCANCODE_FROM_KEYSYM scancode_from_double_keysym106[] = {
/*( / )*/	{ NORM, 0x29,	0x00,	XK_space,		XK_space		},
/*(1/!)*/	{ NORM, 0x16,	0x00,	XK_1,			XK_exclam		},
 /*(2/")*/	{ NORM, 0x1e,	0x00,	XK_2,			XK_quotedbl		},	/*6*/
/*(3/#)*/	{ NORM, 0x26,	0x00,	XK_3,			XK_numbersign	},
/*(4/$)*/	{ NORM, 0x25,	0x00,	XK_4,			XK_dollar		},
/*(5/%)*/	{ NORM, 0x2e,	0x00,	XK_5,			XK_percent		},
 /*(6/&)*/	{ NORM, 0x36,	0x00,	XK_6,			XK_ampersand	},	/*6*/
 /*(7/')*/	{ NORM, 0x3d,	0x00,	XK_7,			XK_apostrophe	},	/*6*/
 /*(8/()*/	{ NORM, 0x3e,	0x00,	XK_8,			XK_parenleft	},	/*6*/
 /*(9/))*/	{ NORM, 0x46,	0x00,	XK_9,			XK_parenright	},	/*6*/
 /*(0/ )*/	{ NORM, 0x45,	0x00,	XK_0,			NoSymbol		},	/*6*/
 /*(-/=)*/	{ NORM, 0x4e,	0x00,	XK_minus,		XK_equal		},	/*6*/
 /*(^/~)*/	{ NORM, 0x55,	0x00,	XK_asciicircum,	XK_asciitilde	},	/*6*/
 /*(\/|)*/	{ NORM, 0x6a,	0x00,	XK_backslash,	XK_bar			},	/*6,106only*/
 /*(H/Z)*/	{ NORM, 0x0e,	0x00,	XK_Hankaku?,	XK_Zenkaku?		},	/*6*/
/*(q/Q)*/	{ NORM, 0x15,	0x14,	XK_q,			XK_Q			},
/*(w/W)*/	{ NORM, 0x1d,	0x00,	XK_w,			XK_W			},
/*(e/E)*/	{ NORM, 0x24,	0x08,	XK_e,			XK_E			},
/*(r/R)*/	{ NORM, 0x2d,	0x00,	XK_r,			XK_R			},
/*(t/T)*/	{ NORM, 0x2c,	0x00,	XK_t,			XK_T			},
/*(y/Y)*/	{ NORM, 0x35,	0x00,	XK_y,			XK_Y			},
/*(u/U)*/	{ NORM, 0x3c,	0x00,	XK_u,			XK_U			},
/*(i/I)*/	{ NORM, 0x43,	0x0c,	XK_i,			XK_I			},
/*(o/O)*/	{ NORM, 0x44,	0x12,	XK_o,			XK_O			},
/*(p/P)*/	{ NORM, 0x4d,	0x13,	XK_p,			XK_P			},
 /*(@/`)*/	{ NORM, 0x54,	0x00,	XK_at,			XK_grave		},	/*6*/
 /*([/{)*/	{ NORM, 0x5b,	0x00,	XK_bracketleft,	XK_braceleft	},	/*6*/
/*(a/A)*/	{ NORM, 0x1c,	0x04,	XK_a,			XK_A			},
/*(s/S)*/	{ NORM, 0x1b,	0x00,	XK_s,			XK_S			},
/*(d/D)*/	{ NORM, 0x23,	0x07,	XK_d,			XK_D			},
/*(f/F)*/	{ NORM, 0x2b,	0x09,	XK_f,			XK_F			},
/*(g/G)*/	{ NORM, 0x34,	0x0a,	XK_g,			XK_G			},
/*(h/H)*/	{ NORM, 0x33,	0x0b,	XK_h,			XK_H			},
/*(j/J)*/	{ NORM, 0x3b,	0x0d,	XK_j,			XK_J			},
/*(k/K)*/	{ NORM, 0x42,	0x0e,	XK_k,			XK_K			},
/*(l/L)*/	{ NORM, 0x4b,	0x0f,	XK_l,			XK_L			},
 /*(;/+)*/	{ NORM, 0x4c,	0x00,	XK_semicolon,	XK_plus			},	/*6*/
 /*(:/*)*/	{ NORM, 0x52,	0x00,	XK_colon,		XK_asterisk		},	/*6*/
 /*(]/})*/	{ NORM, 0x5d,	0x00,	XK_bracketright,XK_braceright	},	/*6*/
/*(z/Z)*/	{ NORM, 0x1a,	0x00,	XK_z,			XK_Z			},
/*(x/X)*/	{ NORM, 0x22,	0x00,	XK_x,			XK_X			},
/*(c/C)*/	{ NORM, 0x21,	0x06,	XK_c,			XK_C			},
/*(v/V)*/	{ NORM, 0x2a,	0x00,	XK_v,			XK_V			},
/*(b/B)*/	{ NORM, 0x32,	0x05,	XK_b,			XK_B			},
/*(n/N)*/	{ NORM, 0x31,	0x11,	XK_n,			XK_N			},
/*(m/M)*/	{ NORM, 0x3a,	0x10,	XK_m,			XK_M			},
/*(,/<)*/	{ NORM, 0x41,	0x00,	XK_comma,		XK_less			},
/*(./>)*/	{ NORM, 0x49,	0x00,	XK_period,		XK_greater		},
/*(//?)*/	{ NORM, 0x4a,	0x00,	XK_slash,		XK_question		},
 /*(_/\)*/	{ NORM, 0x51,	0x00,	XK_underscore,	XK_backslash	},	/*6,106only*/
			{ -1,   -1,		-1,		-1,				-1				},
};
#endif /* 0 */


/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
