/*
 * xuzk - X based frontend for USB eazy keyboard emulator, UZK
 *   by uratan! 2004.5.27
 *     see http://www002.upp.so-net.ne.jp/uratan/UZK/
 */
/*
 * based on 'xev' source codes.
 */
/*

Copyright (c) 1988  X Consortium

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall
not be used in advertising or otherwise to promote the sale, use or
other dealings in this Software without prior written authorization
from the X Consortium.

*/
#include <stdio.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>


/*
 * to decrease the data on PS/2 stream,
 * suppress release scancode when key repeat.
 *   this is for some PS/2-->USB converter, but not so well...
 */
#define DECREASE_LOAD_OF_PS2_STREAM  0

#define WINDOW_MIN_WIDTH  160
#define WINDOW_MIN_HEIGHT  70
#define WINDOW_DEF_WIDTH  160
#define WINDOW_DEF_HEIGHT  70
#define WINDOW_DEF_X 0
#define WINDOW_DEF_Y 0

#ifdef HID_UZK
# define DEF_HID_DEV  "/dev/uhid0"
#else /* HID_UZK */
# ifdef __NetBSD__
#  define DEF_TTY_DEV  "/dev/tty00"
# else /* __NetBSD__ */
#  define DEF_TTY_DEV  "/dev/cuaa0"
# endif /* __NetBSD__ */
#endif /* HID_UZK */

#include "capt_none.xbm"
#include "caption.xbm"
#include "led_none.xbm"
#include "led_off.xbm"
#include "led_on.xbm"
#include "key_off.xbm"
#include "key_on.xbm"
#include "key_off_CA.xbm"
#include "key_on_CA.xbm"
#ifdef HID_UZK
#include "uzk.xbm"
#endif /* HID_UZK */
#include "icon.xbm"

#ifdef HID_UZK
static char *xps2k_version = "xuzk version 0.03";
#else /* HID_UZK */
static char *xps2k_version = "xps2k version 0.25";
#endif /* HID_UZK */

/*
 * functions
 */
int main(int argc, char **argv);
void usage(void);
KeySym do_KeyPressRelease(const char *m, XEvent *eventp);
void set_sizehints(XSizeHints *hintp, int min_width, int min_height, int defwidth, int defheight, int defx, int defy, char *geom);
void init_bitmaps(void);
void clean_bitmaps(void);
void draw_bitmaps(int leds, int keys, int key_CA);
void reset_has_come(int leds);
void leds_have_changed(int leds);

/*
 * variables
 */
int verbose = False;
#ifndef HID_UZK
int reply = False;
int feed_only = False;
#endif /* !HID_UZK */
int CA_control = False;
char *ProgramName;
Display *dpy;
Window win;
int scr;
GC gc;
Pixmap capt_none, led_none;
Pixmap caption;
Pixmap led_off, led_on;
Pixmap key_off, key_on;
Pixmap key_off_CA, key_on_CA;
Pixmap icon;
#ifdef HID_UZK
Pixmap uzk;
#endif /* HID_UZK */
int led_stat, key_stat;
int key_CA_lock;

/*
 * externs
 */
/* in keyconv.c */
extern int convert_and_send_C_A(int pressed);
extern int convert_and_send(KeySym keysym, int pressed);
extern char *make_key_strokes(char *s);

#ifdef HID_UZK
/* in uhid.c */
extern int uhid;		/* file descriptor for uhidX */
extern int hid_read(u_char *buff);
extern void clear_k_m_memory(void);
extern int hid_write(u_char *buff);

#else /* HID_UZK */
/* in ps2.c */
extern int send_reset_done(void);
extern int polling_receive_state(void);
extern int ps2_proto(void);
/* in tty.c */
extern int tty;					/* file descriptor for cuaaX */
extern int initial_lines(void);
extern int receive_a_byte(int do_not_reply);
extern void display_usleep_time(void);
extern void display_wait_time(void);
extern int display_tty_state(void);
extern void tty_test_tool(int mode);
extern int wait_loop;
#endif /* HID_UZK */

#ifdef HID_UZK
extern u_char usb_led[];
extern u_char key_data[];
#endif /* HID_UZK */

/*
 * main entry for xps2k
 */
int main(int argc, char **argv)
{
	char *displayname = NULL;
	char *geom = NULL;
#ifdef HID_UZK
	char *name = "xuzk";
#else /* HID_UZK */
	char *name = "xps2k";
#endif /* HID_UZK */
	int borderwidth = 2;
	Bool reverse = False;
#ifdef HID_UZK
	char *dev = DEF_HID_DEV;
#else /* HID_UZK */
	int initial_ack = False;
	char *dev = DEF_TTY_DEV;
	int stat = False;
	int test = 0;
#endif /* HID_UZK */

	int i;
	XSizeHints hints;
	XSetWindowAttributes attr;
	XWMHints wmhints;
	unsigned long mask;
	int done;
	unsigned long back, fore;
	XGCValues gcv;
	int prev_led_stat;
	char *paste_str_top;
	char *p_paste_str = NULL;
#if DECREASE_LOAD_OF_PS2_STREAM
	int this_is_key_repeat = False;
#endif /* DECREASE_LOAD_OF_PS2_STREAM */

	ProgramName = argv[0];

	/*
	 * analize options
	 */
	for(i=1; i<argc; i++) {
		char *arg = argv[i];
		int wl;

		if(arg[0] != '-')
			usage();

		switch(arg[1]) {
		case 'd':			/* -display host:dpy */
				if(++i >= argc)
					usage();
				displayname = argv[i];
				continue;

		case 'g':			/* -geometry geom */
				if(++i >= argc)
					usage();
				geom = argv[i];
				continue;

		case 'n':			/* -name s */
				if(++i >= argc)
					usage();
				name = argv[i];
				continue;

		case 'b':			/* -bw n */
				if(++i >= argc)
					usage();
				borderwidth = atoi(argv[i]);
				continue;

		case 'r':			/* -rv */
				reverse = True;
				continue;

		case 'l':			/* -l /dev/cuaaX */
				if(++i >= argc)
					usage();
				dev = argv[i];
				continue;

#ifndef HID_UZK
		case 'a':			/* -a */
				reply = True;
				continue;
		case 'f':			/* -f */
				feed_only = True;
				continue;

		case 'i':			/* -i */
				initial_ack = True;
				continue;
#endif /* !HID_UZK */

		case 'c':			/* -c */
				CA_control = True;;
				continue;


		case 'v':			/* -v */
				verbose = True;
				continue;

#ifndef HID_UZK
		case 's':			/* -s */
				stat = True;
				continue;

		case 't':			/* -t, this is hidden option */
				test = 1;
				continue;

		case 'p':			/* -p, this is hidden option */
				test = 2;
				continue;

		case 'w':
				if(++i >= argc)
					usage();
				wl = atoi(argv[i]);
				if((1 <= wl) && (wl <= 300))
					wait_loop = wl;
				else
					usage();
				continue;
#endif /* !HID_UZK */

		default:
				usage();
		}				/* end switch on - */
	}					/* end for over argc */

#ifdef HID_UZK
	/*
	 * try to open USB HID device
	 */
# ifdef __NetBSD__
	uhid = open(dev, O_RDWR);
# else /* __NetBSD__ */
	uhid = open(dev, O_RDWR|O_DIRECT);
# endif /* __NetBSD__ */
	if(uhid < 0) {
		fprintf(stderr, "%s: cannot connect to %s\n", ProgramName, dev);
		goto exit_level_0;
	}
#else /* HID_UZK */
	/*
	 * try to open tty
	 */
# ifdef __NetBSD__
	tty = open(dev, O_RDWR|O_NONBLOCK);
# else /* __NetBSD__ */
	tty = open(dev, O_RDWR|O_DIRECT);
# endif /* __NetBSD__ */
	if(tty < 0) {
		fprintf(stderr, "%s: cannot connect to %s\n", ProgramName, dev);
		goto exit_level_0;
	}

	if(stat) {
		/*
		 * display result of time measurement, then exit
		 */
		display_usleep_time();
		display_wait_time();
		display_tty_state();
		goto exit_level_1;
	}

	/*
	 * make PS/2 CLOCK/DATA as initial
	 */
	initial_lines();

	if(test) {
		/*
		 * generate misc test signal
		 */
		tty_test_tool(test);
		goto exit_level_1;
	}
#endif /* !HID_UZK */

	/*
	 * try to open X server
	 */
	dpy = XOpenDisplay(displayname);
	if(!dpy) {
		fprintf(stderr, "%s: unable to open display '%s'\n",
							ProgramName, XDisplayName(displayname));
		goto exit_level_1;
	}

	/*
	 * arrangements for window operation
	 */
	scr = DefaultScreen(dpy);

	set_sizehints(&hints, WINDOW_MIN_WIDTH, WINDOW_MIN_HEIGHT,
		   WINDOW_DEF_WIDTH, WINDOW_DEF_HEIGHT,
		   WINDOW_DEF_X, WINDOW_DEF_Y, geom);

	if(reverse) {
		back = BlackPixel(dpy, scr);
		fore = WhitePixel(dpy, scr);
	} else {
		back = WhitePixel(dpy, scr);
		fore = BlackPixel(dpy, scr);
	}

	attr.background_pixel = back;
	attr.border_pixel = fore;
	attr.event_mask = ExposureMask |KeyPressMask|KeyReleaseMask|ButtonPressMask;

	mask = CWBackPixel | CWBorderPixel | CWEventMask;

	win = XCreateWindow(dpy, RootWindow(dpy, scr), hints.x, hints.y,
		   hints.width, hints.height, borderwidth, 0,
		   InputOutput, (Visual *)CopyFromParent,
		   mask, &attr);

	init_bitmaps();

	XSetStandardProperties(dpy, win, name, NULL, (Pixmap)0, argv, argc, &hints);

	wmhints.flags = InputHint | IconPixmapHint;
	wmhints.input = True;
	wmhints.icon_pixmap = icon;
		/* if False, you can not enter keys if mouse cursor is on the border */
	XSetWMHints(dpy, win, &wmhints);

	gcv.foreground = fore;
	gcv.background = back;
	gc = XCreateGC(dpy, win, GCForeground | GCBackground, &gcv);
	XSetGraphicsExposures(dpy, gc, False);	/* No need 'NoExpose' event */

	XMapWindow(dpy, win);

	/*
	 *
	 */
	led_stat = 0;		/* bitwised LED status */
	key_stat = 0;		/* how many keys are pushed */
#ifdef HID_UZK
	clear_k_m_memory();
#endif /* HID_UZK */
	key_CA_lock = 0;	/* CTRL+ALT lock, controled by button 2 */

#ifndef HID_UZK
	if(initial_ack)
		send_reset_done();
#endif /* !HID_UZK */

	if(verbose) {
		printf("%s\n", xps2k_version);
		printf("Push button 3 to quit.\n");
	}

	for(done = 0; !done; ) {
		XEvent event;
		KeySym keysym;

#ifdef HID_UZK
		int c;

		prev_led_stat = led_stat;
		c = hid_read(&usb_led[0]);
		if(c > 0) {
			if(usb_led[0] == 'L') {
				led_stat = (usb_led[1] & 0x04/*SCRL*/) ? 0x04 : 0;
				led_stat |= (usb_led[1] & 0x01/*NUML*/) ? 0x01 : 0;
				led_stat |= (usb_led[1] & 0x02/*NUML*/) ? 0x02 : 0;
			}

			if(led_stat != prev_led_stat)
				draw_bitmaps(led_stat, key_stat, key_CA_lock);
		}
#else /* HID_UZK */
		/* does PC send some data for me ? */
		if(polling_receive_state() > 0) {
			if(reply) {
				prev_led_stat = led_stat;
				ps2_proto();
				if(led_stat != prev_led_stat)
					draw_bitmaps(led_stat, key_stat, key_CA_lock);
			} else if(feed_only) {
				int c;
				c = receive_a_byte(/*do_not_reply=*/ 1);
				if(c >= 0) {
					if(verbose)
						printf("0x%02x has come\n", c);
				}
			}
		}
#endif /* HID_UZK */

		/*
		 * is there string to paste ?
		 */
		if(p_paste_str != NULL) {
			KeySym ks;
			int pressed;

			ks = *p_paste_str++;
			pressed = *p_paste_str++;

			if(ks == '\0') {
				p_paste_str = NULL;
				free(paste_str_top);
			} else {
				convert_and_send(ks, pressed);
				/*
				 * PS/2->USB convert may have some limit about transfer rate..
				 */
				usleep(3*1000);
			}

			continue;
		}

		/*
		 * does user want to transport key hit by me ?
		 */
		if(!XEventsQueued(dpy, QueuedAfterFlush)) {
			/*
			 * I expect to wait 3msec (1/3 of standard) for polling interval.
			 *   but
			 * minimum sleep time is influenced by HZ in kernel configuration
			 */
			usleep(3*1000);		/* sleeps about 8ms at HZ=250 kernel */
#if 0
			/*
			 * with arvel AU02-PS, he abondan receive request about 8msec,
			 * xps2k can not capture it. but he can continue if next transmit
			 * has come....
			 * the best resolution is reconfig kernel with HZ=500, but ...
			 * see also wait_for_receive_a_byte() in ps2.c
			 */
#endif /* 0 */
			continue;
		}

		/*
		 * process the XEvent
		 */
		XNextEvent(dpy, &event);

		switch(event.type) {
		case Expose:
				draw_bitmaps(led_stat, key_stat, key_CA_lock);
				break;

		case KeyPress:
				keysym = do_KeyPressRelease("KeyPress", &event);

				convert_and_send(keysym, /*pressed=*/1);

#if DECREASE_LOAD_OF_PS2_STREAM
				if(this_is_key_repeat) {
					this_is_key_repeat = False;
					break;
				}
#endif /* DECREASE_LOAD_OF_PS2_STREAM */

				key_stat++;
				if(key_stat > 5)
					key_stat = 5;

				draw_bitmaps(led_stat, key_stat, key_CA_lock);

#if DECREASE_LOAD_OF_PS2_STREAM
				this_is_key_repeat = False;
#endif /* DECREASE_LOAD_OF_PS2_STREAM */
				break;

		case KeyRelease:

#if DECREASE_LOAD_OF_PS2_STREAM
				/*
				 * at key repeat, Release+Press events may be queued
				 * in a pair, so suppress some jobs and PS/2 stream
				 */
				this_is_key_repeat = False;

				if(XEventsQueued(dpy, QueuedAfterFlush)) {
					XEvent event_followed;

					XNextEvent(dpy, &event_followed);

					if(
						(event_followed.type == KeyPress) &&
						(event.xkey.keycode == event_followed.xkey.keycode)
					) {
						this_is_key_repeat = True;
					}

					XPutBackEvent(dpy, &event_followed);
				}

				if(this_is_key_repeat) {
					/*
					 * PC can sense next press scancode as PRESS,
					 * no need to send release scancode.
					 */
					break;
				}
#endif /* DECREASE_LOAD_OF_PS2_STREAM */

				keysym = do_KeyPressRelease("KeyRelease", &event);

				convert_and_send(keysym, /*pressed=*/0);

				key_stat--;
				if(key_stat < 0)
					key_stat = 0;

				draw_bitmaps(led_stat, key_stat, key_CA_lock);
				break;

		case ButtonPress:
				if(((XButtonEvent *)&event)->button == 3) {
					/*
					 * quit this program
					 */
					if(verbose)
						printf("Quit by Button 3\n");
					done = 1;
				} else if(((XButtonEvent *)&event)->button == 2) {
					if(CA_control) {
						/*
						 * reverse CTRL+ALT lock
						 */
						key_CA_lock = 1 - key_CA_lock;

						if(key_CA_lock)
							convert_and_send_C_A(/*pressed=*/1);
						else
							convert_and_send_C_A(/*pressed=*/0);

						draw_bitmaps(led_stat, key_stat, key_CA_lock);
						if(verbose)
							printf("reverse CTRL+ALT lock\n");
					} else {
						char *line;
						int nbytes;

						line = XFetchBytes(dpy, &nbytes);

						if((verbose) && (nbytes != 0))
								printf("\"%s\" pasted\n", line);

						paste_str_top = make_key_strokes(line);
						if(paste_str_top != NULL) {
							p_paste_str = paste_str_top;
						}
						XFree(line);
					}
				} else {
					/*
					 * reset pushed-key graphics
					 */
					key_stat = 0;
#ifdef HID_UZK
					clear_k_m_memory();
					hid_write(key_data);
#endif /* HID_UZK */
					draw_bitmaps(led_stat, key_stat, key_CA_lock);
					if(verbose)
						printf("redraw keys\n");
				}
				break;

		default:
				break;
		}
	}

	/*XUngrabPointer(dpy, CurrentTime);*/

	/*
	 * release about X connection
	 */
	clean_bitmaps();
	XCloseDisplay(dpy);

#ifndef HID_UZK
	/*
	 * release about tty
	 */
	initial_lines();
#endif /* !HID_UZK */

exit_level_1:
#ifdef HID_UZK
	close(uhid);
#else /* HID_UZK */
	close(tty);
#endif /* HID_UZK */

exit_level_0:
	exit(0);
}

/*
 * display usage and quit
 */
void usage(void)
{
	static char *msg[] = {
		"-d display     X server to contact",
		"-g geometry    size and location of window",
		"-n string      window name",
		"-b pixels      border width in pixels",
		"-r             reverse video",
#ifdef HID_UZK
		"-l /dev/uhidX  USB keyboard bridge to connect",
#else /* HID_UZK */
# ifdef __NetBSD__
		"-l /dev/tty0X  tty to connect",
# else /* __NetBSD__ */
		"-l /dev/cuaaX  tty to connect",
# endif /* __NetBSD__ */
		"-a             reply for request",
		"-f             only feed the request",
		"-i             enable initial ack",
#endif /* HID_UZK */
		"-c             enable CTRL+ALT control by mouse",
		"-v             verbose mode",
#ifndef HID_UZK
		"-s             only display statistic",
		"-w number      give wait_loop value ([1,300])",
#endif /* !HID_UZK */
		NULL
	};
	char **cpp;

	fprintf(stderr, "%s\n", xps2k_version);
	fprintf(stderr, "  usage: %s [-options ...]\n", ProgramName);
	fprintf(stderr, "    where options include:\n");

	for(cpp = msg; *cpp; cpp++)
		fprintf(stderr, "      %s\n", *cpp);

	exit(1);
}

/*
 * display about X key event info
 */
KeySym do_KeyPressRelease(const char *m, XEvent *eventp)
{
	XKeyEvent *e = (XKeyEvent *)eventp;
	KeySym ks;
	char *ksname;

	XLookupString(e, NULL, 0, &ks, NULL);

	if(verbose) {
		if(ks == NoSymbol)
			ksname = "NoSymbol";
		else if(!(ksname = XKeysymToString(ks)))
			ksname = "(no name)";

		printf("%s - keycode %u (%s)", m, e->keycode, ksname);
	}

	return(ks);
}

/*
 * display about X key event info
 */
void set_sizehints(XSizeHints *hintp, int min_width, int min_height, int defwidth, int defheight, int defx, int defy, char *geom)
{
	int geom_result;

	/* set the size hints, algorithm from xlib xbiff */

	hintp->width = hintp->min_width = min_width;
	hintp->height = hintp->min_height = min_height;
	hintp->flags = PMinSize;
	hintp->x = hintp->y = 0;
	geom_result = NoValue;

	if(geom != NULL) {
		geom_result = XParseGeometry(geom, &hintp->x, &hintp->y,
						(unsigned int *)&hintp->width,
						(unsigned int *)&hintp->height);
		if((geom_result & WidthValue) && (geom_result & HeightValue)) {
#define max(a,b) ((a) > (b) ? (a) : (b))
			hintp->width = max(hintp->width, hintp->min_width);
			hintp->height = max(hintp->height, hintp->min_height);
			hintp->flags |= USSize;
		}

		if((geom_result & XValue) && (geom_result & YValue)) {
			hintp->flags += USPosition;
		}
	}

	if(!(hintp->flags & USSize)) {
		hintp->width = defwidth;
		hintp->height = defheight;
		hintp->flags |= PSize;
	}

	if(geom_result & XNegative) {
		hintp->x = DisplayWidth(dpy, DefaultScreen(dpy)) + hintp->x -
			hintp->width;
	}
	if(geom_result & YNegative) {
		hintp->y = DisplayHeight(dpy, DefaultScreen(dpy)) + hintp->y -
			hintp->height;
	}

	return;
}

/*
 * create pixmaps
 */
void init_bitmaps(void)
{
	capt_none = XCreateBitmapFromData(dpy, win,
				capt_none_bits, capt_none_width, capt_none_height);
	led_none = XCreateBitmapFromData(dpy, win,
				led_none_bits, led_none_width, led_none_height);
	caption = XCreateBitmapFromData(dpy, win,
				caption_bits, caption_width, caption_height);
	led_off = XCreateBitmapFromData(dpy, win,
				led_off_bits, led_off_width, led_off_height);
	led_on = XCreateBitmapFromData(dpy, win,
				led_on_bits, led_on_width, led_on_height);
	key_off = XCreateBitmapFromData(dpy, win,
				key_off_bits, key_off_width, key_off_height);
	key_on = XCreateBitmapFromData(dpy, win,
				key_on_bits, key_on_width, key_on_height);
	key_off_CA = XCreateBitmapFromData(dpy, win,
				key_off_CA_bits, key_off_CA_width, key_off_CA_height);
	key_on_CA = XCreateBitmapFromData(dpy, win,
				key_on_CA_bits, key_on_CA_width, key_on_CA_height);
	icon = XCreateBitmapFromData(dpy, win,
				icon_bits, icon_width, icon_height);
#ifdef HID_UZK
	uzk = XCreateBitmapFromData(dpy, win,
				uzk_bits, uzk_width, uzk_height);
#endif /* HID_UZK */

	return;
}

/*
 * destroy pixmaps
 */
void clean_bitmaps(void)
{
	XFreePixmap(dpy, caption);
	XFreePixmap(dpy, led_off);
	XFreePixmap(dpy, led_on);
	XFreePixmap(dpy, key_off);
	XFreePixmap(dpy, key_on);
	XFreePixmap(dpy, key_off_CA);
	XFreePixmap(dpy, key_on_CA);
	XFreePixmap(dpy, icon);
#ifdef HID_UZK
	XFreePixmap(dpy, uzk);
#endif /* HID_UZK */

	return;
}

/*
 * draw window
 */
void draw_bitmaps(int leds, int keys, int key_CA)
{
	int i;
	int xx, yy;
	Pixmap p;

	xx = 47;
	yy =  3;

	XClearArea(dpy, win, xx, yy, caption_width, caption_height, False);

#ifdef HID_UZK
	p = caption;
#else /* HID_UZK */
	if(reply)
		p = caption;
	else
		p = capt_none;
#endif /* HID_UZK */

	XCopyPlane(dpy, p, win, gc, 0, 0,
								caption_width, caption_height, xx, yy, 1);

	/*
	 * draw 3 LEDs
	 */
	for(i=0; i<3; i++) {
		xx = 47 + 1 + i * 22;
		yy =  3 + 12;

#ifdef HID_UZK
		if(leds & (1<<i))
			p = led_on;
		else
			p = led_off;
#else /* HID_UZK */
		if(reply) {
			if(leds & (1<<i))
				p = led_on;
			else
				p = led_off;
		} else {
			p = led_none;
		}
#endif /* HID_UZK */

		XClearArea(dpy, win, xx, yy, led_off_width, led_off_height, False);

		XCopyPlane(dpy, p, win, gc, 0, 0,
									led_off_width, led_off_height, xx, yy, 1);
	}

	/*
	 * draw 5 key tops
	 */
	for(i=0; i<5; i++) {
		xx =  2 + key_off_width * i;
		yy =  3 + 12 + 20;

		if((CA_control) && (i==4)) {
			if(key_CA)
				p = key_on_CA;
			else
				p = key_off_CA;
		} else {
			if(i < keys)
				p = key_on;
			else
				p = key_off;
		}

		XClearArea(dpy, win, xx, yy, key_off_width, key_off_height, False);

		XCopyPlane(dpy, p, win, gc, 0, 0,
									key_off_width, key_off_height, xx, yy, 1);
	}

#ifdef HID_UZK
	xx = 118;
	yy = 2;
	XClearArea(dpy, win, xx, yy, uzk_width, uzk_height, False);
	XCopyPlane(dpy, uzk, win, gc, 0, 0, uzk_width, uzk_height, xx, yy, 1);
#endif /* HID_UZK */

	return;
}

/*
 * called when PK_RESET_KBD has come
 */
void reset_has_come(int leds)
{
	led_stat = leds;
	key_stat = 0;
	key_CA_lock = 0;

	draw_bitmaps(led_stat, key_stat, key_CA_lock);
	XFlush(dpy);

	return;
}

/*
 * called when PK_SET_LEDS has come
 */
void leds_have_changed(int leds)
{
	led_stat = leds;		/* bitwised LED status */

	draw_bitmaps(led_stat, key_stat, key_CA_lock);
	XFlush(dpy);

	return;
}
